from enum import Enum
from typing import MutableMapping, MutableSequence
import serial
import os

from .BluetoothDevice import BluetoothDevice
from .BluetoothLink import BluetoothLink, A2dpLink, AvrcpLink, HfpLink
from .HfpState import HfpState
from .AvrcpState import AvrcpState
from .BluetoothRole import BluetoothRole
from .BluetoothProfile import BluetoothProfile
from .BluetoothCodec import BluetoothCodec
from .BluetoothCallStatus import BluetoothCallStatus
from .CurrentCodecs import CurrentCodecs
from .OnOff import OnOff
from .SerialMessage import SerialMessage, payload_identifier_lookup
from .exceptions import *
from .DeviceModel import DeviceModel

### code from pyserial.tools.list_ports
if os.name == 'nt':
    from serial.tools.list_ports_windows import comports
elif os.name == 'posix':
    from serial.tools.list_ports_posix import comports
else:
    raise ImportError("Sorry: no implementation for your platform ('{}') available".format(os.name))
###

def strip_response(response_string):
    return response_string.split('=')[1].split(':')[0].strip()

class PtdBtDevice:
    def __init__(self, model_number: str, max_name_length: int, max_payload_length: int) -> None:

        self._serial_port = None

        self.MAX_BT_FRIENDLY_NAME_LENGTH = max_name_length
        self.MAX_HOST_PAYLOAD_LEN_BYTES = max_payload_length

        # port iteration code from pyserial.tools.list_ports
        iterator = comports(include_links=True)
        for n, (port, desc, hwid) in enumerate(iterator, 1):
            if desc.startswith( f"PTD {model_number.upper()}" ):
                self._serial_port = serial.Serial(port)

        if self._serial_port == None:
            raise DeviceNotFoundException;

    def __del__(self) -> None:
        # if we're connected to a serial port, close it.
        if self._serial_port != None:
            self._serial_port.close()

    def _send_packet(self, type: SerialMessage, payload: str) -> None:
        if len(payload) > self.MAX_HOST_PAYLOAD_LEN_BYTES:
            raise PayloadLengthExceedsLimitsException
        self._serial_port.reset_input_buffer()
        self._serial_port.reset_output_buffer()
        self._serial_port.write(f"#{payload_identifier_lookup[type]}{chr(ord('0') + len(payload))}{payload}:".encode('utf-8'))

    def _read_packet(self) -> str:
        packet = self._serial_port.read_until(b":").decode()
        return packet

    def get_firmware_version(self) -> str:
        message_type = SerialMessage.MSG_GET_BT_FW_VERSION
        payload = ""
        self._send_packet(message_type, payload)
        fw_string = self._read_packet()
        return strip_response(fw_string)

    def get_model(self) -> DeviceModel:
        message_type = SerialMessage.MSG_GET_CHASSIS_MODEL_NUMBER
        payload = ""
        self._send_packet(message_type, payload)
        model_number_string = self._read_packet()
        return DeviceModel[strip_response(model_number_string).replace("-","_")]

    def get_serial_number(self) -> int:
        message_type = SerialMessage.MSG_GET_CHASSIS_ESN
        payload = ""
        self._send_packet(message_type, payload)
        serial_number_string = self._read_packet()
        return int(strip_response(serial_number_string))

    def get_manufacture_date(self) -> str:
        message_type = SerialMessage.MSG_GET_MFG_DATE
        payload = ""
        self._send_packet(message_type, payload)
        mfg_date_string = self._read_packet()
        return strip_response(mfg_date_string)

    def restore_defaults(self) -> None:
        message_type = SerialMessage.MSG_RESTORE_FACTORY_DEFAULTS
        payload = ""
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def save_defaults(self) -> None:
        message_type = SerialMessage.MSG_SAVE_DEFAULTS
        payload = ""
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def set_bt_name(self, name: str) -> None:
        if name > self.MAX_BT_FRIENDLY_NAME_LENGTH:
            raise FriendlyNameLengthExceedsLimitsException
        message_type = SerialMessage.MSG_SET_BT_NAME
        payload = name
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def get_bt_name(self) -> str:
        message_type = SerialMessage.MSG_GET_BT_NAME
        payload = ""
        self._send_packet(message_type, payload)
        bt_name_string = self._read_packet()
        return strip_response(bt_name_string)

    def get_bt_address(self) -> str:
        message_type = SerialMessage.MSG_GET_BT_ADDR
        payload = ""
        self._send_packet(message_type, payload)
        bt_addr_string = self._read_packet()
        return strip_response(bt_addr_string)

    def perform_bt_inquiry(self, inquiry_time_sec: int) -> MutableSequence[BluetoothDevice]:
        message_type = SerialMessage.MSG_START_BT_INQUIRY
        payload = str(chr(inquiry_time_sec))
        self._send_packet(message_type, payload)
        bt_device_list = []
        response = ""
        while response != "#TransactionDone:":
            response = self._read_packet()
            if response.startswith("#BtDev"):
                mac_address = response.split("=")[1].split(",")[0]
                friendly_name = response.split(",")[1].split(":")[0]
                bt_device_list.append(BluetoothDevice(mac_address, friendly_name))
        return bt_device_list

    def get_detected_devices(self) -> MutableSequence[BluetoothDevice]:
        message_type = SerialMessage.MSG_GET_DETECTED_DEVICES_LIST
        payload = ""
        self._send_packet(message_type, payload)
        bt_device_list = []
        response = ""
        while response != "#TransactionDone:":
            response = self._read_packet()
            if response.startswith("#BtDev"):
                mac_address = response.split("=")[1].split(",")[0]
                friendly_name = response.split(",")[1].split(":")[0]
                bt_device_list.append(BluetoothDevice(mac_address, friendly_name))
        return bt_device_list

    def start_bt_connection(self, bt_address: str, bt_profile: BluetoothProfile) -> BluetoothCallStatus:
        message_type = SerialMessage.MSG_START_BT_CONNECTION
        payload = bt_address
        payload += str(bt_profile.value)
        self._send_packet(message_type, payload)
        call_status_string = self._read_packet()
        return BluetoothCallStatus[f"CALL_STATUS_{strip_response(call_status_string).upper()}"]

    def disconnect_all(self) -> None:
        message_type = SerialMessage.MSG_DISCONNECT_ALL
        payload = ""
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def get_bt_role(self) -> BluetoothRole:
        message_type = SerialMessage.MSG_GET_BT_ROLE
        payload = ""
        self._send_packet(message_type, payload)
        bt_role_string = self._read_packet()
        return BluetoothRole[strip_response(bt_role_string)]

    def set_bt_profile_codec_pair(self, profile: BluetoothProfile, codec: BluetoothCodec) -> None:
        message_type = SerialMessage.MSG_SET_PROFILE_CODEC
        payload = str(profile.value) + str(codec.value)
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device
        bt_profiles_and_codecs_array = self.get_bt_profiles_and_codecs()
        if profile == BluetoothProfile.A2DP:
            if bt_profiles_and_codecs_array[0][1] != codec:
                raise InvalidBluetoothConfigurationException
        else:
            if bt_profiles_and_codecs_array[1][1] != codec:
                raise InvalidBluetoothConfigurationException

    def get_current_bt_codecs(self) -> CurrentCodecs:
        message_type = SerialMessage.MSG_GET_PROFILE_CODEC_LIST
        payload = ""
        self._send_packet(message_type, payload)
        profiles_and_codecs_string = self._read_packet().split("=")[1]
        a2dp_codec = None
        hfp_codec = None
        for profile_codec_pair in profiles_and_codecs_string.split(","):
            profile = BluetoothProfile[profile_codec_pair.split("<")[1].split(";")[0]]
            codec = BluetoothCodec[profile_codec_pair.split(";")[1].split(">")[0]]
            if profile == BluetoothProfile.A2DP:
                a2dp_codec = codec
            else:
                hfp_codec = codec
        return CurrentCodecs(a2dp_codec, hfp_codec)

    def get_open_bt_links(self) -> MutableSequence[BluetoothLink]:
        message_type = SerialMessage.MSG_GET_OPEN_LINK_LIST
        payload = ""
        self._send_packet(message_type, payload)
        open_link_list = []
        response = ""
        while response != "#TransactionDone:":
            response = self._read_packet()
            if response.startswith("#BtLink"):
                response = strip_response(response)
                profile = BluetoothProfile[response.split(" ")[3]]
                mac_address = response.split(" ")[4]
                current_link = None
                if profile == BluetoothProfile.A2DP:
                    codec = BluetoothCodec[response.split(" ")[5]]
                    sampling_rate = int(response.split(" ")[6])
                    current_link = A2dpLink(mac_address, codec, sampling_rate)
                elif profile == BluetoothProfile.HFP:
                    codec = BluetoothCodec.WIDE_BAND
                    if response.split(" ")[5] == "NB":
                        codec = BluetoothCodec.NARROW_BAND
                    hfp_state = HfpState[response.split(" ")[6]]
                    current_link = HfpLink(mac_address, codec, hfp_state)
                elif profile == BluetoothProfile.AVRCP:
                    avrcp_state = AvrcpState[response.split(" ")[5]]
                    current_link = AvrcpLink(mac_address, avrcp_state)
                open_link_list.append(current_link)
        return open_link_list

    def get_paired_device_list(self) -> MutableSequence[str]:
        message_type = SerialMessage.MSG_GET_PAIR_LIST
        payload = ""
        self._send_packet(message_type, payload)
        paired_address_list = []
        response = ""
        while response != "#TransactionDone:":
            response = self._read_packet()
            if response.startswith("#BtPair"):
                paired_address_list.append(strip_response(response))
        return paired_address_list

    def clear_paired_device_list(self) -> None:
        message_type = SerialMessage.MSG_CLEAR_PAIR_LIST
        payload = ""
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def set_a2dp_audio_state(self, on_or_off: OnOff) -> None:
        message_type = SerialMessage.MSG_SET_A2DP_AUDIO_STATE
        payload = str(on_or_off.value)
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def set_hfp_audio_state(self, on_or_off: OnOff) -> None:
        message_type = SerialMessage.MSG_SET_HFP_AUDIO_STATE
        payload = str(on_or_off.value)
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def get_bt_firmware_version(self) -> str:
        message_type = SerialMessage.MSG_GET_BT_FW_VERSION
        payload = ""
        self._send_packet(message_type, payload)
        bt_fw_version_string = self._read_packet()
        return strip_response(bt_fw_version_string)

    def reset_bt(self) -> None:
        message_type = SerialMessage.MSG_RESET_BT
        payload = ""
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def set_bt_command_timeout(self, timeout: int) -> None:
        message_type = SerialMessage.MSG_SET_BT_COMMAND_TIMEOUT
        payload = str(chr(timeout))
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def get_bt_command_timeout(self) -> int:
        message_type = SerialMessage.MSG_GET_BT_COMMAND_TIMEOUT
        payload = ""
        self._send_packet(message_type, payload)
        bt_command_timeout_string = self._read_packet()
        return int(strip_response(bt_command_timeout_string))

    def set_avrcp_volume(self, volume: int) -> None:
        message_type = SerialMessage.MSG_SET_AVRCP_VOLUME
        payload = str(chr(volume))
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def get_avrcp_volume(self) -> int:
        message_type = SerialMessage.MSG_GET_AVRCP_VOLUME
        payload = ""
        self._send_packet(message_type, payload)
        avrcp_volume_string = self._read_packet()
        return int(strip_response(avrcp_volume_string))

    def set_hfp_volume(self, volume: int) -> None:
        message_type = SerialMessage.MSG_SET_HFP_VOLUME
        payload = str(chr(volume))
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def get_hfp_volume(self) -> int:
        message_type = SerialMessage.MSG_GET_HFP_VOLUME
        payload = ""
        self._send_packet(message_type, payload)
        hfp_volume_string = self._read_packet()
        return int(strip_response(hfp_volume_string))

    def set_mitm_state(self, on_or_off: OnOff) -> None:
        message_type = SerialMessage.MSG_SET_MITM_STATE
        payload = str(on_or_off.value)
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def get_mitm_state(self) -> OnOff:
        message_type = SerialMessage.MSG_GET_MITM_STATE
        payload = ""
        self._send_packet(message_type, payload)
        mitm_state_string = self._read_packet()
        return OnOff[strip_response(mitm_state_string)]

    def set_bt_pin(self, pin: int) -> None:
        message_type = SerialMessage.MSG_SET_BT_PIN
        payload = str(pin)
        self._send_packet(message_type, payload)
        self._read_packet() # read the TransactionDone packet back from the device

    def get_bt_pin(self) -> int:
        message_type = SerialMessage.MSG_GET_BT_PIN
        payload = ""
        self._send_packet(message_type, payload)
        bt_pin_string = self._read_packet()
        return int(strip_response(bt_pin_string))

    def get_bt_link_key(self) -> str:
        message_type = SerialMessage.MSG_GET_BT_LINK_KEY
        payload = ""
        self._send_packet(message_type, payload)
        bt_link_key_string = self._read_packet()
        return strip_response(bt_link_key_string)

    def get_rssi(self) -> int:
        message_type = SerialMessage.MSG_GET_RSSI
        payload = ""
        self._send_packet(message_type, payload)
        rssi_value_string = self._read_packet()
        return int(strip_response(rssi_value_string))
